﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  public class LMConstantsOptimizer {
    private bool ApplyLinearScaling { get; set; }
    private int MaximumIterations { get; set; }

    public LMConstantsOptimizer() {

    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="tree">The tree for which the constants are optimized.</param>
    /// <param name="nodes">The nodes which should be adapted. The nodes must be the same (reference) as the ones used in the tree and the double values specify the inita</param>
    /// <param name="x">The input data.</param>
    /// <param name="y">The targer date.</param>
    /// <param name="applyLinearScaling">Flag that determines whether linear scaling nodes should be added during the optimization.</param>
    /// <returns> Fit of the symbolic expression tree in terms of ... </returns>
    public static double OptimizeConstants(ISymbolicExpressionTree tree,IRegressionProblemData problemData, IEnumerable<int> rows, bool applyLinearScaling, int maxIterations = 10) {
      //if (tree == null) throw new ArgumentNullException("tree");
      //if (nodes == null) throw new ArgumentNullException("nodes");
      //if (initialConstants == null) throw new ArgumentNullException("intitialConstants");
      //if (problemData == null) throw new ArgumentNullException("problemData");

      //if (!nodes.Any()) return 0;

      var ds = problemData.Dataset;
      var variables = ConstantsOptimization.Util.GenerateVariables(ds);
      var laggedVariables = ConstantsOptimization.Util.ExtractLaggedVariables(tree);
      var allVariables = variables.Union(laggedVariables);

      double[,] x = ConstantsOptimization.Util.ExtractData(ds, allVariables, rows);
      double[] y = ds.GetDoubleValues(problemData.TargetVariable, rows).ToArray();

      double[] initialConstants;

      AutoDiff.IParametricCompiledTerm term;
      if (!TreeToAutoDiffTermConverter.TryConvertToAutoDiff(tree, applyLinearScaling, allVariables, out term, out initialConstants))
        throw new NotSupportedException("Could not optimize constants of symbolic expression tree due to not supported symbols used in the tree.");

      var constants = (double[])initialConstants.Clone();
      alglib.lsfitstate state;
      alglib.lsfitreport rep;
      int retVal;

      int numberOfRows = x.GetLength(0);
      int numberOfColumns = x.GetLength(1);
      int numberOfConstants = constants.Length;

      try {
        alglib.lsfitcreatefg(x, y, constants, numberOfRows, numberOfColumns, numberOfConstants, cheapfg: false, state: out state);
        alglib.lsfitsetcond(state, 0.0, 0.0, maxIterations);
        //alglib.lsfitsetgradientcheck(state, 0.001);
        alglib.lsfitfit(state, Evaluate, EvaluateGradient, null, term);
        alglib.lsfitresults(state, out retVal, out constants, out rep);
      } catch (ArithmeticException) {
        return double.NaN;
      } catch (alglib.alglibexception) {
        return double.NaN;
      }

      ConstantsOptimization.Util.UpdateConstants(tree, constants);

      return rep.r2;
    }

    private static void Evaluate(double[] c, double[] x, ref double fx, object o) {
      AutoDiff.IParametricCompiledTerm term = (AutoDiff.IParametricCompiledTerm)o;
      fx = term.Evaluate(c, x);
    }

    private static void EvaluateGradient(double[] c, double[] x, ref double fx, double[] grad, object o) {
      AutoDiff.IParametricCompiledTerm term = (AutoDiff.IParametricCompiledTerm)o;
      Tuple<double[], double> result = term.Differentiate(c, x);
      fx = result.Item2;
      Array.Copy(result.Item1, grad, grad.Length);
    }

    public static bool CanOptimizeConstants(ISymbolicExpressionTree tree) {
      return TreeToAutoDiffTermConverter.IsCompatible(tree);
    }
  }
}
